/* The software provided here is released by the National
 * Institute of Standards and Technology (NIST), an agency of
 * the U.S. Department of Commerce, Gaithersburg MD 20899,
 * USA.  The software bears no warranty, either expressed or
 * implied. NIST does not assume legal liability nor
 * responsibility for a User's use of the software or the
 * results of such use.
 * Please note that within the United States, copyright
 * protection, under Section 105 of the United States Code,
 * Title 17, is not available for any work of the United
 * States Government and/or for any works created by United
 * States Government employees. User acknowledges that this
 * software contains work which was created by NIST employees
 * and is therefore in the public domain and not subject to
 * copyright.  The User may use, distribute, or incorporate
 * this software provided the User acknowledges this via an
 * explicit acknowledgment of NIST-related contributions to
 * the User's work. User also agrees to acknowledge, via an
 * explicit acknowledgment, that any modifications or
 * alterations have been made to this software before
 * redistribution.
 * --------------------------------------------------------------------
 *
 * Change History:
 * Simson L. Garfinkel - May 1, 2008
 *   - Major rewrite using new function
 * Simson L. Garfinkel - simsong@acm.org - August 21, 2006
 *   - Re-implemented based on Douglas White's original Perl code.
 *   - First 4096-bytes of file describes parameters
 *   - Single executable for both 128-bit and 160-bit Bloom filters.
 *   - Uses memmap() for handling 512MB bloom filters. 
 * Douglas White - douglas.white@nist.gov - June 21, 2003
 *   Original implementation in perl.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <errno.h>
#include <err.h>
#include <ctype.h>

#include "bloom.h"

#define BLOOM_VECTOR_OFFSET 4096	/* size of one page */

static int *hexcharvals = 0;
int nsrl_klookup_counters[32];	/* keep track of the number of times 0, 1, 2 functions used for lookup */

/** Initialization function is used solely for hex output
 * for debugging functions.
 */
static void nsrl_bloom_init()
{
    if(hexcharvals==0){
	/* Need to initialize this */
	int i;
	hexcharvals = calloc(sizeof(int),256);
	for(i=0;i<10;i++){
	    hexcharvals['0'+i] = i;
	}
	for(i=10;i<16;i++){
	    hexcharvals['A'+i-10] = i;
	    hexcharvals['a'+i-10] = i;
	}
	memset(nsrl_klookup_counters,0,sizeof(nsrl_klookup_counters));
    }
}


/**
 * Convert a hex representation to binary, and return
 * the number of bits converted.
 * @param binbuf output buffer
 * @param hex    input buffer (in hex)
 */
int nsrl_hex2bin(unsigned char *binbuf,const char *hex)
{
    int bits = 0;
    if(hexcharvals==0) nsrl_bloom_init();
    while(hex[0] && hex[1]){
	*binbuf++ = ((hexcharvals[(unsigned char)hex[0]]<<4) |
		     hexcharvals[(unsigned char)hex[1]]);
	hex  += 2;
	bits += 8;
    }
    return bits;
}

/**
 * Convert a binary blob to a hex representation
 */

const char *nsrl_hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int flag)
{
    int charcount = 0;
    const char *start = dst;		// remember where the start of the string is
    const char *fmt = (flag & NSRL_HEXBUF_UPPERCASE) ? "%02X" : "%02x";

    if(hexcharvals==0) nsrl_bloom_init();
    *dst = 0;				// begin with null termination
    while(bytes>0 && dst_len > 3){
	int add_spaces = 0;

	sprintf(dst,fmt,*bin); // convert the next byte
	dst += 2;
	bin += 1;
	dst_len -= 2;
	bytes--;
	charcount++;			// how many characters
	
	if(flag & NSRL_HEXBUF_SPACE2) add_spaces = 0;
	if((flag & NSRL_HEXBUF_SPACE4) && charcount%2==0){
	    *dst++ = ' ';
	    *dst   = '\000';
	    dst_len -= 1;
	}
    }
    return start;			// return the start
}


/**
 * Sets a numbered bit of the bloom filter.
 */
static inline void set_bloom_bit(nsrl_bloom *b,unsigned int bit)
{
    int q = 1<<(bit%8);
    b->vector[bit/8] |= q;
}

/** Gets a numbered bit of the bloom filter.
 */
static inline int get_bloom_bit(const nsrl_bloom *b,unsigned int bit)
{
    int q = 1<<(bit%8);
    return (b->vector[bit/8] & q) ? 1 : 0;
}


/**
 * Converts bloom filter parameters into ascii for storage in filter.
 */
void nsrl_bloom_info(char *buf,int buflen, const nsrl_bloom *b)
{
    snprintf(buf,buflen,
	     "nsrl_bf_version:2\n"
	     "hash_bits:%d\n"
	     "bloom_bits:%d\n"
	     "k:%d\n"
	     "added_items:%d\n"
	     "comment:%s\n",
	     b->hash_bits, b->M,b->k,b->added_items, b->comment);
}

/**
 * Histogram analysis for printing information about usage of bloom filter.
 */
void nsrl_calc_histogram(const nsrl_bloom *b,unsigned int counts[256])
{
    int i;
    unsigned int byteswbits = 0;

    memset(counts,0,sizeof(counts));
    for(i=0;i<b->vector_bytes;i++){
    	if(b->vector[i]) {
	    counts[b->vector[i]]++;
	    byteswbits++;}
    }
}

void nsrl_print_histogram(const nsrl_bloom *b,const unsigned int counts[256])
{
    unsigned int totalbits = 0;
    unsigned int bitcount = 0;
    int i,j,mask;
    float top, bottom, percent = 0.0;
    for(i=1;i<256;i++){
	if(counts[i] > 0){
	    for(j = 0; j<8; j++){
		mask = (1 << j);
		if(i&mask){ bitcount += counts[i]; } 
	    }
	}
    }
    if(b->M < 32) totalbits = 1<<b->M;
    else{ totalbits--;}
    printf("%u bits total, %u bits set\n",totalbits, bitcount);
    top = (float)bitcount;
    bottom = (float)totalbits;
    percent = top/bottom;
    percent *= 100;
    printf("%f percent full\n", percent);
}


static int nsrl_bloom_info_update(nsrl_bloom *b)
{
    if(b->fd==0) return 0;		/* no open file to update */
    char buf[BLOOM_VECTOR_OFFSET];
    memset(buf,0,sizeof(buf));
    nsrl_bloom_info(buf,sizeof(buf),b);
    if (lseek(b->fd,0,SEEK_SET)) return -1;
    if (write(b->fd,buf,BLOOM_VECTOR_OFFSET)!=BLOOM_VECTOR_OFFSET) return -1;
    return 0;
}

/* Return the ith bit of a blob */
static inline unsigned int get_bit(const unsigned char *buf,unsigned int bit)
{
    return (buf[bit/8] & (1<<bit%8)) ? 1 : 0;
}


/*
 * Encrypt with HMAC of hash; store the results in ebuf.
 */
static inline const unsigned char *hash_encrypt(const nsrl_bloom *b,const unsigned char *hash,unsigned char *ebuf)
{
    unsigned len = b->hash_bytes;	/* make openSSL happy */
    HMAC(b->md,b->key,b->hash_bytes,hash,b->hash_bytes,ebuf,&len);
    return ebuf;
}

/**
 * Add a hash into the bloom filter.
 */
void nsrl_bloom_add(nsrl_bloom *b,const unsigned char *hash)
{
    /* Split the hash k times according to bloom_bits(M) */
    unsigned int i,j;
    u_char ebuf[20];
    if (b->key) hash = hash_encrypt(b,hash,ebuf);
    if (b->debug>1) { printf("VECTORS:"); }
    for(i=0;i<b->k;i++){
	unsigned int offset = i * b->M;
	unsigned int v = 0;
	for(j=0;j<b->M;j++){
	    v = (v<<1) | get_bit(hash,offset+j);
	}
	if(b->debug>1) { printf("Setting bit: %u.\n",v); }
	set_bloom_bit(b,v);
	if(b->debug>1) { printf("\t%08x",v);if(i%3==2) putchar('\n');}
    }
    b->added_items++;
    if (b->added_items % 1000==0) nsrl_bloom_info_update(b);
    if (b->debug>1) { printf("\n"); }
}

void nsrl_bloom_addString(nsrl_bloom *b,const char *str)
{
    u_char buf[EVP_MAX_MD_SIZE];
    unsigned int buflen = EVP_MAX_MD_SIZE;
    EVP_Digest((const u_char *)str,strlen(str),buf,&buflen,b->md,0);
    nsrl_bloom_add(b,buf);
}

/* nsrl_bloom_query:
 * Check each round of nist_function128 in the vector. If any are not set,
 * then the hash is not in the bloom filter.
 *
 * Returns: 1 if present, 0 if not present
 */
int nsrl_bloom_query(const nsrl_bloom *b,const unsigned char *hash)
{
    u_int i,j;
    u_char ebuf[20];
    if (b->debug){
	char buf[1024];
	printf("nsrl_bloom_query(%s)\n",nsrl_hexbuf(buf,sizeof(buf),hash,b->hash_bytes,0));
    }
    if (b->key) hash = hash_encrypt(b,hash,ebuf);
    for(i=0;i<b->k;i++){		/* i is which vector function to query */
	nsrl_klookup_counters[i]++;		/* increment the  */

	u_int offset = i * b->M;
	u_int v = 0;		/* v is the bit in the vector that is to be queried */
	for(j=0;j<b->M;j++){
	    v = (v<<1) | get_bit(hash,offset+j);
	}
	if(b->debug>1){ printf("  %s V(0x%08x)=%d\n",
			       (i==0?"VECTORS":"       "),
			       v,get_bloom_bit(b,v)); }
	if(get_bloom_bit(b,v)==0) return 0; /* if that bit wasn't set, return early */
    }
    /* All of the bits were set; hash must be in the bloom filter */
    if(b->debug>1) putchar('\n');
    return 1;				/* must be present */
}

int nsrl_bloom_queryString(const nsrl_bloom *b,const char *str)
{
    u_char buf[EVP_MAX_MD_SIZE];
    u_int buflen = EVP_MAX_MD_SIZE;
    EVP_Digest((const u_char *)str,strlen(str),buf,&buflen,b->md,0);
    return nsrl_bloom_query(b,buf);
}


/****************************************************************
 *** Routines for creating and opening bloom filters.
 ****************************************************************/

nsrl_bloom *nsrl_bloom_alloc()
{
    nsrl_bloom *b = (nsrl_bloom *)calloc(sizeof(*b),1);
    b->free_this = 1;
    return b;
}

static void nsrl_bloom_set_params(nsrl_bloom *b)
{
    b->hash_bytes = b->hash_bits/8;
    switch(b->hash_bytes){
    case 16:
	b->md = EVP_md5();
	break;
    case 20:
	b->md = EVP_sha1();
	break;
    default:
	fprintf(stderr,"nsrl_bloom_set_params: hash_bytes=%d?\n",b->hash_bytes);
	exit(1);
    }
    b->key = malloc(b->hash_bytes);
}


#ifndef O_BINARY
#define O_BINARY 0
#endif


#define xstr(s) str(s)
#define str(s) #s
int nsrl_bloom_open(nsrl_bloom *b,const char *fname,int mode)
{
    char offset_buf[BLOOM_VECTOR_OFFSET];
    char *line,*buf;
    int version = 0;
    int  prot = 0;

    switch(mode & O_ACCMODE){
    case O_RDONLY: prot = PROT_READ;if(b->debug) printf("PROT_READ\n");break;
    case O_WRONLY: prot = PROT_WRITE;if(b->debug) printf("PROT_WRITE\n");break;
    case O_RDWR:   prot = PROT_READ|PROT_WRITE;if(b->debug) printf("PROT_READ|PROT_WRITE\n");break;
    }

    /* Open the file and get the parameters */

    b->fd = open(fname,mode|O_BINARY,0);
    if(b->fd<0) return -1;			/* could not open */

    if(read(b->fd,offset_buf,sizeof(offset_buf))!=sizeof(offset_buf)){
	errno = EINVAL;
	return -1;			/* not big enough for header? */
    }
    buf = offset_buf;
    
    while((line = strsep(&buf,"\n"))){
	char *colon = index(line,':');
	if(colon){
	    *colon = 0;			/* terminate at the colon */
	    colon++;
	    if(strcmp(line,str(nsrl_bf_version))==0) version = atoi(colon);
	    if(strcmp(line,str(hash_bits))==0)       b->hash_bits = atoi(colon);
	    if(strcmp(line,str(bloom_bits))==0)      b->M = atoi(colon);
	    if(strcmp(line,str(k))==0)               b->k = atoi(colon);
	    if(strcmp(line,str(added_items))==0)     b->added_items = atoi(colon);
	    if(strcmp(line,str(comment))==0)         b->comment = strdup(colon);
	}
    }
    nsrl_bloom_set_params(b);

    if(version!=2){
	fprintf(stderr,"bloom: require nsrl bf vesion 2; got version %d\n",version);
	errno= EINVAL;
	return -1;
    }
    if(b->hash_bits==0 || b->M==0 || b->k==0){
	close(b->fd);
	errno= EINVAL;
	fprintf(stderr,"bloom: invalid parameters. hash_bits=%d bloom_bits=%d k=%d\n",
		b->hash_bits,b->M,b->k);
	return -1;
    }
    b->vector_bytes  = 1 << (b->M-3);
    b->vector = mmap(0,b->vector_bytes, prot, MAP_FILE|MAP_SHARED, b->fd,BLOOM_VECTOR_OFFSET);
    if(b->vector == MAP_FAILED){
	return -1;			/* ugh. */
    }
    b->memmapped = 1;
    return 0;
}

/** nsrl_bloom_create:
 * @param fname - name of the filter to create; 0 if in-memory
 * @param hash_bits  - the number of bits in values to be hashed.
 * @param bloom_bits - log2(m)
 * @param k = k
 * @param comment - a comment to store in the bloom filter.
 */

int nsrl_bloom_create(nsrl_bloom *b,
		      const char *fname,int hash_bits, int bloom_bits,int k,
		      const char *comment)
{
    b->vector_bytes = 1 << (bloom_bits - 3);	/* Needed vector size */
    b->vector = calloc(b->vector_bytes,1);
    if(b->vector==0) return -1;			/* not enough memory to allocate vector buf */

    if(hash_bits < bloom_bits * k) err(1,"bloom_bits * k > hash_bits (%d * %d > %d)",bloom_bits,k,hash_bits);
    if(b->vector_bytes < (1<<(bloom_bits-3))) err(1,"vector isn't big enough (%d * 8 < 2^(%d-1))",
					       b->vector_bytes,bloom_bits);

    b->hash_bits     =  hash_bits;
    b->M             =  bloom_bits;
    b->k             =  k;
    b->comment       =  strdup(comment);
    nsrl_bloom_set_params(b);


    if(fname==0) return 0;		/* nameless bloom filter; we are done */

    /* Create the file with the paremters and empty bloom filter */
    b->fd = open(fname,O_CREAT|O_RDWR|O_EXCL|O_BINARY,0666); 
    if(b->fd<0){
	return -1;		/* failed */
    }

    if(nsrl_bloom_info_update(b)){	/* try to write the parameters */
	unlink(fname);			/* erase the file */
	return -1;			/* some writing problem */
    }
    /* Tab out to where the vector starts.
     * We pre-allocate the file (avoid making it sparse) to prevent the file being
     * fragmented on the hard drive.
     */
    lseek(b->fd,BLOOM_VECTOR_OFFSET,SEEK_SET); 
    if(write(b->fd,b->vector,b->vector_bytes)!=b->vector_bytes){ /* write failure? */
	unlink(fname);
	return -1;			/* malloc problem */
    }
    /* Now map the bloom filter into memory */
    free(b->vector);
    b->vector = mmap(0,b->vector_bytes,PROT_READ|PROT_WRITE,MAP_FILE|MAP_SHARED,b->fd,BLOOM_VECTOR_OFFSET);
    b->memmapped = 1;
    return 0;
}

/**
 * Establish a passphrase for the specified bloom filter.
 */

void nsrl_set_passphrase(nsrl_bloom *b,const char *passphrase)
{
    EVP_Digest((const u_char *)passphrase,strlen(passphrase),b->key,&b->hash_bytes,b->md,0);
}

void nsrl_bloom_clear(nsrl_bloom *b)
{
    if(b->vector && (b->vector!=MAP_FAILED)){
	if(b->memmapped) munmap(b->vector,b->vector_bytes);
	else free(b->vector);		/* free the memory otherwise */
    }
    if(b->fd)      close(b->fd);	/* close the file */
    if(b->comment) free(b->comment);
    if(b->key){
	memset(b->key,0,b->hash_bytes);
	free(b->key);
    }
    memset(b,0,sizeof(*b));		/* clean object reuse */
}

void nsrl_bloom_free(nsrl_bloom *b)
{
    nsrl_bloom_clear(b);
    free(b);
}
