/*
 * frag_find: given a file, find all of its fragments on the hard drive.
 * Automatically ignore popular fragments. 
 */

#include "bloom.h"
#include "bloomconfig.h"
#include "aftimer.h"

#define __STDC_FORMAT_MACROS

#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <signal.h>
#include <string.h>
#include <cstring>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#if defined(HAVE_SYS_DISK_H) 
#include <sys/disk.h>
#endif
#if defined(HAVE_SYS_MOUNT_H)
#include <sys/mount.h>
#endif


#include <map>
#include <vector>

#define USE_BITSET
#ifdef USE_BITSET
#include <bitset>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifndef DEFAULT_BLOCKSIZE
#define DEFAULT_BLOCKSIZE 512
#endif

#ifdef HAVE_ERR_H
#include <err.h>
#else
void err(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n",strerror(errno));
  exit(eval);
}
#endif

#ifndef HAVE_ATOI64
int64_t atoi64(const char *buf)
{
    int64_t ret=0;
    sscanf(buf,"%"PRId64,&ret);
    return ret;
}
#endif


u_int blocksize = DEFAULT_BLOCKSIZE;

uint64_t blocknumber=0;
void status(int n)
{
    printf("current block: %"PRId64"\n",blocknumber);
}

/* We keep a map of SHA1s to block numbers, so that a page is not processed twice */
class sha1_t {
public:
    u_char buf[20];
    /** hexbuf returns the hex representation of this string.
     * Not that this is not threadsafe.
     */     
    const char *hex() const{
	static char hexbuf[41];
	for(int i=0;i<20;i++){
	    sprintf(hexbuf+i*2,"%02x",buf[i]);
	}
	return hexbuf;
    }
    bool operator<(const sha1_t &s2) const
    {
	return memcmp(this->buf,s2.buf, 20) < 0;
    }
};

struct lsha1
{
    bool operator()(struct sha1_t s1, struct sha1_t s2) const
    {
	return memcmp(s1.buf,s2.buf, 20) < 0;
    }
    bool operator()(struct sha1_t *s1, struct sha1_t *s2) const
    {
	return memcmp(s1->buf,s2->buf, 20) < 0;
    }
};



/** blocklist is a list of block numbers.
 */

class blocklist_t:public std::vector<int64_t> {
public:
    /** return true if the blocklist contains a given block number */
    bool contains(int64_t block) const {
	for(blocklist_t::const_iterator i = this->begin();
	    i != this->end();
	    i++){
	    if(*i == block) return true;
	}
	return false;
    }
};

/* A filemap is an array of arrays. The key is the block number of the mapped file.
 * the value is an array of all the locations where the block appars
 */
class filemap_t:public std::vector<blocklist_t> {
public:
    /* Clean the filemap. If any sector in the input file is in more than
     * one place in the image, check to see if the previous next is in just one place.
     * If so, then only pick the sector that aligns.
     */
    void clean(){
	for(int64_t b=1;b<(int64_t)this->size();b++){
	    if((*this)[b-1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b-1].front()+1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
	
	/* Clean in the reverse direction */
	for(int64_t b=this->size()-1;b>=0;b--){
	    if((*this)[b+1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b+1].front()-1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
    }

    /* Return the length of a run --- the number of sequential blocks
     * starting at a give location.
     */
    int64_t runlen(int64_t b){
	int64_t len = 1;
	while((b+1 < (int64_t)this->size()) &&
	      ((*this)[b+1].contains((*this)[b].front()+1))){
	    len++;
	    b++;
	}
	return len;
    }

    int64_t found_count(){
	int64_t count = 0;
	for(int64_t b=0;b<(int64_t)this->size();b++){
	    if ((*this)[b].size()>0) count++;
	}
	return count;
    }

    double found_percent(){
	return found_count() * 100.0 / this->size();
    }

    void print(){
	printf("Target Block(s)     Found at image block\n");
	for(int64_t b=0;b<(int64_t)this->size();){
	    int64_t runlen = this->runlen(b);

	    if(runlen>1){
		printf("%10"PRId64"-%-10"PRId64"    %10"PRId64"-%-10"PRId64"   (%"PRId64" blocks)\n",
		       b,b+runlen-1,
		       (*this)[b].front(),
		       (*this)[b].front()+runlen-1,
		       runlen);
	    }
	    else{
		bool first=true;
		for(blocklist_t::const_iterator i = (*this)[b].begin();
		    i != (*this)[b].end();
		    i++){
		    if(first){
			printf("%10"PRId64"               ",b);
			first = false;
		    }
		    printf(" %10"PRId64" ",*i);
		}
		if(first==false) printf("\n");
	    }
	    b += runlen;
	}
    }



};



/** sha1map maps individual sha1's to a list of blocks
 */

class sha1map_t:public std::map<class sha1_t,blocklist_t> {
};





/** blockfile is a simple abstraction that lets you read a file block-by-block.
 */

class blockfile {
    int    fd;
    struct stat st;
    static const int bufsize=65536;
    u_char buf[bufsize];
    int64_t block_at_buffer0;
    ssize_t bytes_in_buffer;
public:
    u_int blocksize;
    const char *filename;
    uint64_t filesize;
    uint64_t blocks;			// number of blocks
    blockfile():fd(0){}
    /** open the file and return the fd, or return <0 if error. */
    int open(const char *fname,int blocksize){
	this->filename  = fname;
	this->blocksize = blocksize;
	fd = ::open(fname,O_RDONLY | O_BINARY,0666);
	if(fd>0) {
	    if(::fstat(fd,&st)<0){
		::close(fd);
		fd = 0;
		return -1;
	    }
	    this->filesize = st.st_size;
	    this->blocks   = this->filesize / this->blocksize;
	    if(this->filesize==0 && (((st.st_mode & S_IFMT)==S_IFBLK) ||
				     ((st.st_mode & S_IFMT)==S_IFCHR))){
		/* Figure out the size of a block device */
#ifdef __APPLE__
		int sector_size = 0;
		uint64_t total_sectors = 0;
		uint64_t max_read_blocks = 0;
		if(ioctl(fd,DKIOCGETBLOCKSIZE,&sector_size)){
		    sector_size = 512;		// assume 512
		}
		if(ioctl(fd,DKIOCGETBLOCKCOUNT,&total_sectors)){
		    total_sectors=0;		// seeking not allowed on stdin
		}
		if(ioctl(fd,DKIOCGETMAXBLOCKCOUNTREAD,&max_read_blocks)){
		    max_read_blocks = 0;	// read all you want
		}
		this->blocks   = total_sectors;
		this->filesize = sector_size * total_sectors;
#endif
#if defined(__FreeBSD__) && defined(DIOCGSECTORSIZE)
		int sector_size = 0;
		uint64_t total_sectors = 0;
		uint64_t max_read_blocks = 0;
		if(ioctl(fd,DIOCGSECTORSIZE,&sector_size)){
		    sector_size = 512;		// can't figure it out; go with the default
		}
		off_t inbytes=0;
		if(ioctl(fd,DIOCGMEDIASIZE,&inbytes)){
		    total_sectors = 0;
		}
		this->blocks = inbytes / sector_size;
		this->filesize = inbytes;
#endif
#ifdef linux
#ifdef BLKGETSIZE64
		uint64_t total_bytes=0;
		if(ioctl(fd,BLKGETSIZE64,&total_bytes)){
		    total_bytes = 0;
		}
#else
		int total_bytes=0;
		if(ioctl(fd,BLKGETSIZE,&total_bytes)){
		    total_bytes = 0;
		}
#endif
		this->filesize = total_bytes;
		this->blocks   = total_bytes / BLOCK_SIZE;
#endif
	    }
	}
	loadblock(0);			// preload with first
	return fd;
    }
    ~blockfile(){
	if(fd>0) ::close(fd);
    }
    /** Return a block from the disk image. Does caching. 
     * Returns the number of bytes read, or <0 for error.
     */
    void loadblock(off_t blocknumber){
	off_t pos = blocknumber * blocksize;
	lseek(fd,pos,SEEK_SET);
	bytes_in_buffer = ::read(fd,buf,bufsize);
	block_at_buffer0 = blocknumber;
    }

    const u_char *getblock(off_t blocknumber){
	/* See if we need to reload the buffer */
	if(blocknumber<block_at_buffer0 ||
	   blocknumber>=(block_at_buffer0+bufsize/blocksize)){
	    loadblock(blocknumber);
	}
	return buf + (blocknumber-block_at_buffer0)*blocksize;
    }
};


void usage()
{
    printf("usage: frag_find [options] <target> <raw disk file>\n");
    printf("\n");
    printf("options:\n");
    printf("     -s <start>   - start image scan at block <start> (default is beginning)\n");
    printf("     -e <end>     - end image scan at block <start> (default is end)\n");
    printf("     -b blocksize - sets the blocksize (default is %d)\n",DEFAULT_BLOCKSIZE);
    printf("     -r           - prints the raw association map, in addition to the cleaned on\n");
    exit(1);
}


/** return a very fast-to-calculate 24-bit value from a 512-byte buffer*/
uint32_t val3char(const u_char *buf)
{
    uint32_t counter = 0;
    for(int i=0;i<512;i++){
	counter = ((counter << 1) | (counter>>31)); // rotate and Xor-in a bit
	counter = counter ^ (counter >> 8 & 0x1) ^ buf[i];
    }
    return counter & 0xffffff;
}

int main(int argc,char **argv)
{
    int opt_raw = 0;
    uint64_t opt_start = 0;
    uint64_t opt_end   = 0;
    int ch;
    while ((ch = getopt(argc,argv,"s:e:rb:h?")) != -1){
	switch(ch){
	case 's': opt_start = atoi64(optarg);break;
	case 'e': opt_end   = atoi64(optarg);break;
	case 'r': opt_raw   = 1;break;
	case 'b': blocksize = atoi(optarg); break;
	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    argc -= optind;
    argv += optind;
    
    if(argc<2) usage();

    if(blocksize<4){
	printf("Blocksizes smaller than 4 are meaningless. Try again.\n");
	exit(1);
    }

#ifdef SIGINFO
    signal(SIGINFO,status);
#endif

    blockfile target,imagefile;

    if(target.open(*argv++,blocksize)<0) err(1,"Cannot open %s",target.filename);
    if(imagefile.open(*argv++,blocksize)<0) err(1,"Cannot open %s",imagefile.filename);

    if(target.filesize < blocksize){
	fprintf(stderr,"Image size (%"PRId64") is smaller than blocksize (%d); try again.\n",
		target.filesize,blocksize);
	exit(1);
    }
    if(imagefile.filesize==0){
	fprintf(stderr,"Image size is zero?\n");
	exit(1);
    }
    if(opt_end==0) opt_end = imagefile.blocks;


    printf("Scanning for %s (%"PRId64" %d-byte blocks) in %s (%"PRId64" blocks)\n",
	   target.filename,target.blocks,blocksize,imagefile.filename,imagefile.blocks);

    printf("Creating bloom filter...\n");
    NSRLBloom b;
    b.create(0,160,32,5,"In-memory bloom filter for frag find");
    printf("Bloom filter created.\n");
    printf("Storing SHA1 values of search blocks in bloom filter...\n");

    sha1map_t sha1map;

#ifdef USE_BITSET
    
    std::bitset<16777216> seenfirst3;	// bitset to hold whether or not first three seen
#endif    

    /** Read through the target file block by block and create a SHA1 structure
     * for each block that is read. Each SHA1 must be allocated because we need a place
     * to keep all of these SHA1s. The map points to each SHA1.
     */
    for(blocknumber=0;blocknumber < target.blocks; blocknumber++){
	sha1_t *sha1 = new sha1_t();
	const u_char *buf = target.getblock(blocknumber);
	SHA1((u_char *)buf,blocksize,sha1->buf); // make the sha1

	// report if it is in the input file twice
	if(opt_raw && b.query(sha1->buf)){
	    printf("Input file block %"PRId64" (%s) appears previously at ",blocknumber,sha1->hex());
	    for(blocklist_t::const_iterator i = sha1map[*sha1].begin();
		i!=sha1map[*sha1].end();
		i++){
		printf("%"PRId64" ",*i);
	    }
	    printf("\n");
	}

	// add to the map and to the bloom filter
	sha1map[*sha1].push_back(blocknumber);
	b.add(sha1->buf);

#ifdef USE_BITSET
	// Finally, record that we have seen a block beginning with these 3 characters
	seenfirst3[val3char(buf)] = 1;
#endif
    }
    printf("All SHA1 values have been stored\n");

    printf("Now searching image file...\n");
    uint64_t count = 0;

    // Allocate an array to hold all of the block lists
    filemap_t filemap;
    filemap.resize(target.blocks);

    /** Read through the image file block by block. This time we can use a shared SHA1 structure
     * because we do not need to allocate a hash for each block.
     */

    aftimer timer;
    timer.start();
    int in=0,out=0;
    for(uint64_t blocknumber=opt_start;blocknumber < opt_end && blocknumber < imagefile.blocks; blocknumber++){
	struct sha1_t sha1;

	if(blocknumber>opt_start && (((blocknumber-opt_start) % 100000)==0)){
	    int64_t blocks = blocknumber-opt_start;
	    int64_t total = opt_end-opt_start;
	    printf("%"PRId64"M out of %"PRId64"M sectors processed; hits=%"PRId64,
		   blocks/1000000,total/1000000,count);
	    if(timer.elapsed_seconds()>1.0){

		int left = timer.eta((double)blocks/(double)total) / 60;
		printf("; %3g Kblocks/sec; done in %d days, %d hours, %d min",
		       (blocks/1000)/timer.elapsed_seconds(),
		       left / (24*60),
		       (left / 60) % 24,
		       left % 60);
		printf("; in=%d out=%d ",in,out);
	    }
	    putchar('\n');
	    fflush(stdout);
	}

	/* Scan through the input file and the SHA1 of each block */
	const u_char *buf = imagefile.getblock(blocknumber);

#ifdef USE_BITSET
	u_int v3 = val3char(buf);
	//int   res = seenfirst3[v3];
	if(blocknumber>50000){
	    //printf("buf[0]=%u buf[1]=%u buf[2]=%u\n",buf[0],buf[1],buf[2]);
	    //printf("blocknumber=%qd seenfirst3[%d]=%d in=%d out=%d\n",blocknumber,v3,res,in,out);
	    if(seenfirst3[v3]==0){
		out++;
		continue;
	    }
	    in++;
	}
#endif

	SHA1((u_char *)buf,blocksize,sha1.buf);

	/* If the block's SHA1 is in the bloom filter, get the list of blocks 
	 * where that SHA1 was observed and each one to the filemap.
	 */

	if(b.query(sha1.buf)){
	    if(opt_raw) printf("Block %"PRId64" appears to contain the input file\n",blocknumber);
	    if(opt_raw) printf("   >>> found at ");
	    for(blocklist_t::const_iterator i = sha1map[sha1].begin();
		i!=sha1map[sha1].end();
		i++){
		if(opt_raw) printf("   %"PRId64" ",*i);
		filemap[*i].push_back(blocknumber);
	    }
	    if(opt_raw) printf("\n");
	    count++;
	}
    }

    printf("\n\n");
    printf("Blocksize: %d\n",target.blocksize);
    printf("Target file: %s  (%"PRId64" blocks)\n",target.filename,target.blocks);
    printf("Image file:  %s  (%"PRId64" blocks)\n",imagefile.filename,imagefile.blocks);
    if(opt_start!=0){
	printf("   NOTE: Image scan started at block %"PRId64"\n",opt_start);
    }
    if(opt_end!=imagefile.blocks){
	printf("   NOTE: Image scan ended at block %"PRId64"\n",opt_end);
    }
    printf("Blocks of target file found in image file: %"PRId64"\n",count);
    printf("Here is where they were found:\n");

    if(opt_raw){
	printf("RAW FILEMAP:\n");
	filemap.print();
	printf("===============================================================================\n");
	printf("\n");
    }

    filemap.clean();
    /* note: perhaps it would make more sense to iteratively clean. */
    filemap.print();
    printf("Total blocks of original file found: %"PRId64" (%2.0f%%)\n", filemap.found_count(),filemap.found_percent());

    return 0;
}
